/*
 * LowBeam.c
 * 
 * Электронное реле ближнего света.
 *
 * Author: Погребняк Дмитрий, г. Самара, 2014
 *
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей,
 * включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 */ 

#include "LowBeam.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

uint8_t current_state;
uint16_t countdown;

inline uint_fast8_t is_ignition_on() {
  return !(IGNITION_PIN & IGNITION);
}

inline uint_fast8_t is_switch_on() {
  return (SWITCH_PIN & SWITCH);
}

// Прерывание по таймеру, выполняется 2353 раза в секунду
ISR(TIM0_OVF_vect) {
  if (countdown > 0) {
    countdown--;
    return;
  }
  switch (current_state) {
    case STATE_SLOW_WARM: { // медленный прогрев
      uint8_t o = PWM_OC;
      if (o >= 254) {
        PWM_OC = 255;
        current_state = STATE_ON;
      } else {     
        o++;
        PWM_OC = o;
        if (o < SLOW_RAMP_VALUE) {
          countdown = SLOW_RAMP_SPEED;
        } else if (o == SLOW_RAMP_VALUE) {
          countdown = SLOW_RAMP_HOLD;
        } else if (o < SLOW_WARM_VALUE) {
          countdown = SLOW_WARM_SPEED;
        } else {
          countdown = SLOW_LAST_SPEED;
        }
      }      
    } break;
    case STATE_FAST_WARM: { // быстрый прогрев
      uint8_t o = PWM_OC;
      if (o >= 254) {
        PWM_OC = 255;
        current_state = STATE_ON;
      } else {      
        o++;
        PWM_OC = o;
        if (o < FAST_WARM_VALUE) {
          countdown = FAST_WARM_SPEED;
        } else {
          countdown = FAST_LAST_SPEED;
        }
      }      
    } break;
    case STATE_HOLD: { // удержание после отключения
      uint8_t o = PWM_OC;
      if (o <= 1) {
        PWM_OC = 0;
        current_state = STATE_LIGHT_OFF;
      } else {
        o--;
        PWM_OC = o;
        countdown = HOLD_FALL_SPEED;
      }
    } break;
  }
}

// Прерывание по любому изменению логических уровней на входах "зажигание" и "выключатель света"
ISR(PCINT0_vect) {
  if (is_ignition_on()) { 
    if (is_switch_on()) {
      // Если зажигание включено, и выключатель включен.
      switch (current_state) {
        case STATE_IGNITION_OFF: 
          current_state = STATE_SLOW_WARM; 
          countdown = 0; 
          break;
        case STATE_HOLD:
        case STATE_LIGHT_OFF:
          current_state = STATE_FAST_WARM;
          countdown = 0;
          break;
      }
    } else {
      // Если зажигание включено, но выключатель отключен
      switch (current_state) {
        case STATE_SLOW_WARM:
          if (PWM_OC >= HOLD_VALUE) {
            PWM_OC = HOLD_VALUE;
            current_state = STATE_HOLD;
            countdown = HOLD_COUNTDOWN;
          } else {
            current_state = STATE_IGNITION_OFF;
            PWM_OC = 0;
            countdown = 0;
          }
          break;
        case STATE_FAST_WARM: 
          current_state = STATE_HOLD;
          if (PWM_OC >= HOLD_VALUE) {
            PWM_OC = HOLD_VALUE;
            countdown = HOLD_COUNTDOWN;
          } else {
            countdown = 0;
          }
          break;
        case STATE_ON:
          PWM_OC = HOLD_VALUE;
          current_state = STATE_HOLD;
          countdown = HOLD_COUNTDOWN;
          break;
      }
    }
  } else { // Если зажигание выключено - то выкл, без вариантов.
    current_state = STATE_IGNITION_OFF;
    PWM_OC = 0;
  }
}

int main(void)
{
  DDRB = 0;
  PORTB = (1 << PORTB2) | (1 << PORTB0); // pull-up на свободных портах, чтобы не болтались
  // Конфигурируем порты
  PWM_PORT |= PWM; // oc0b;
  PWM_DDR |= PWM;
  
  SWITCH_PORT &= ~SWITCH;
  SWITCH_DDR &= ~SWITCH;
  
  IGNITION_DDR &= ~IGNITION;
  IGNITION_PORT |= IGNITION;

  // конфигурируем таймер
  PCMSK = (1 << PCINT3) | (1 << PCINT4); // Прерывание по изменению пина по входам 3 и 4
  GIMSK |= (1 << PCIE); // Включение прерывания по изменению пина
  
  PWM_OC = 0;
  PWM_DDR &= ~PWM; // отключение выхода PWM, иначе при первом цикле таймера на нём будет низкий уровень
  TIMSK0 = 0;
  TIFR0 |= (1 << OCF0B);
  
  TCCR0A = 0b00110001; // OC0B - инверсный выход, режим phase correct PWM
  TCCR0B = 0b00000001; // прескалер 1 к 1
  
  while (!(TIFR0 & (1 << OCF0B))); // ждём прохождения первого цикла 
  TIFR0 |= (1 << TOV0) | (1 << OCF0B);
  PWM_DDR |= PWM; // включение выхода PWM
  
  TIMSK0 |= (1 << TOIE0); // разрешение прерывания по переполнению таймера
  
  // Начальное чтение пинов, если сразу включено и зажигание и лампы, то переход к стадии "разогрев"
  
  countdown = 0;
  if (is_ignition_on() && is_switch_on()) {
    current_state = STATE_SLOW_WARM;
  } else {
    current_state = STATE_IGNITION_OFF;
  }
  
  set_sleep_mode(SLEEP_MODE_IDLE);
  sleep_enable();
  sei();
  while(1) {
    sleep_cpu();
  }
}